:- module(_,_,
%[main/1, term2latex/2, pred2texcommand/2, list2delim/4],
	[fsyntax, assertions]).

:- use_module(library(compiler),[use_module/1]).
:- use_module(library(aggregates)).
:- use_module(library(terms)).

:- function(arith(false)). %  not needed with fsyntax

:- comment(author, "Edison Mera").

:- comment(module, "This program convert prolog goals to latex entries.

A term is converted in a latex text using a special semantic, that is:

@item The atoms or numbers are preserved.

@item A var is converted to latex as \\_.

@item The lists of terms are converted in lists of latex entries.

@item If a term goal(Goal, Pattern) is found, all solutions for Goal
  are obtained and all the values for Pattern are converted to latex.
  This let us automatically to obtain list of prolog data in latex
  formats.

@item If the term unifies with a predicate which is a separator, then
  each argument is converted, and in the generated latex text, it
  appears separated with the specified separator.

@item Each predicate will be converted to a latex command entry, and
  the arguments of the predicate are used as the latex command
  arguments.

In case you are not comfortable with the format, you must change your
style sheets or main latex file, and maintain this program generic,
because this could be used by other programs.  There is NOT NECESSARY
to change this program.

").

:- multifile pred_db/1.

main(Args) :-
	process_args(Args).

help_message :=
'
  Usage:
    pl2latex [-h] | -i FromFile1 [-i FromFile2|-i FromFile3 ...] -|ToFile

    This program convert prolog goals to latex entries.

'.

process_args(['-h'|Args]) :-
	display(user_error, ~help_message),
	process_args(Args).

process_args(['-i', FromFile|Args]) :-
	use_module(FromFile),
	!,
	process_args(Args).

process_args([LatexFile]) :-
	LatexFile == '-' ->
	write_latex_stream
 ;
	write_latex_file(LatexFile).

process_args([]) :-
	display_string(~help_message).

write_latex_file(LatexFile) :-
	open(LatexFile, write, O),
	current_output(CO),
	set_output(O),
	write_latex_stream,
	set_output(CO),
	close(O).

display_list(Stream, List) :-
	current_output(CO),
	set_output(Stream),
	display_list(List),
	set_output(CO).

write_latex_stream :-
	pred_db(A),
	term2latex(A,L),
	display(user_error,'{NOTE: generated latex commands are:\n'),
	pred2texcommand(L,C),
	display(user_error, C),
	display(user_error,'}\n').

pred2texcommand([],[]).
pred2texcommand([F/N|L], [X|C]) :-
	atom_number(AN,N),
	atom_concat(['\\newcommand{\\', F, '}[', AN, ']\n'], X),
	pred2texcommand(L,C).

:- data commandName/1.

term2latex(A,L) :-
	retractall_fact(commandName(_)),
	term2latex4(A,0,_),
	findall(X,commandName(X),L),
	retractall_fact(commandName(_)).

list2latex([], _I).
list2latex([A|As], I) :-
	term2latex1(A, I),
	list2latex(As, I).

list2latex_nb([], _I).
list2latex_nb([A|As], I) :-
	term2latex_nb(A, I),
	list2latex_nb(As, I).

% At this point put the identation and new line
term2latex1(A,I) :-
	identation(I),
	term2latex2(A,I),
	nl.

% Same as term2latex1 without brackets
term2latex_nb(A,I) :-
	identation(I),
	term2latex3(A,I,NeedId),
	(   NeedId == 1 ->
	    true
	;   nl
	).

% At this point put brackets
term2latex2(A,I) :-
	display('{'),
	term2latex3(A,I, NeedId),
	(   NeedId == 1 ->
	    identation(I)
	;   true
	),
	display('}').

term2latex3(A,I, NeedId) :-
	I1 is I + 1,
	term2latex4(A,I1, NeedId).

term2latex4(A,I, NeedId) :-
	(   (atom(A) ; num(A)) ->
	    display(A),
	    NeedId = 0
	;   var(A) ->
	    display('\\_'),
	    NeedId = 0
	;   list(A) ->
	    nl,
	    list2latex(A,I),
	    NeedId = 1
	;   A = goal(Goal, Pattern) ->
	    nl,
	    (   call(Goal),
		term2latex_nb(Pattern, I),
		fail
	    ;	NeedId = 1
	    )
	;   A =.. [F|Args] ->
	    (   separator(F, SepText) ->
		list2seplatex(Args,SepText,I)
	    ;	display_list(['\\',F,'\n']),
		list2latex(Args,I),
		(   functor(A, F, N) ->
		    (   commandName(F/N) ->
			true
		    ;   asserta_fact(commandName(F/N))
		    )
		;   true
		),
		NeedId = 1
	    )
	).

list2seplatex([],        _SepText, _I).
list2seplatex([Arg],     _SepText,  I) :-
	term2latex4(Arg, I, _),!.
list2seplatex([Arg|Args], SepText,  I) :-
	term2latex4(Arg, I, _),
	display(SepText),!,
	list2seplatex(Args, SepText, I).

%separator(_Name, _Text).
separator(',', ', ').
separator(and, ' y ').
separator(or,  ' o ').
separator('-', ' - ').
separator(';', '; ').
separator('+', ' + ').
separator('-', ' - ').

:- comment(list2delim(List, Delim, DelimEnd, Term), "Unifies Term with
   a term that contains the elements of @var{List} delimited with
   @var{Delim}, and the last element delimited with @var{DelimEnd}").

list2delim([A],   _Delim, _DelimEnd, A) :- !.
list2delim([A,B], _Delim,  DelimEnd, C) :-
	C =.. [DelimEnd,A,B],
	!.
list2delim([A|As], Delim,  DelimEnd, C) :-
	C =.. [Delim, A, Cs],
	!,
	list2delim(As, Delim, DelimEnd, Cs).

identation(N) :-
	N > 0 ->
	display('  '),
	N2 is N - 1,
	identation(N2)
 ;
	true.
