%% :- module(support,[sort_papers/3]).

:- use_package([assertions]).

%% CIAO Specific
:- use_module(library(format)).
:- use_module(library(dynamic),[assert/1]).

:- comment(module,"This library contains miscellaneous support
   predicates.").

:- push_prolog_flag(multi_arity_warnings,off).

%-----------------------------------------------------------------------------
:- pred butype_description(Butype,Text) # "@var{Text} is the textual
   description of @var{Butype} in English. This description should be
   short (will probably be used as a section title).".
%-----------------------------------------------------------------------------

butype_description(A,B) :- butype_description(A,eng,B).

%-----------------------------------------------------------------------------
:- pred butype_description(Butype,Lang,Text) # "@var{Text} is the
   textual description of @var{Butype} in language @var{Lang}. This
   description should be short (will probably be used as a section
   title).".
%-----------------------------------------------------------------------------

%% 1 * Refereed/serious publications:
%% ----------------------------------
%% - Only refereed articles published in a journal:
butype_description("article",eng,"Articles in Refereed Journals").
butype_description("article",esp,"Art\\'{\\i}culos en Revistas").
%% - Only refereed articles published in conference proceedings. 
butype_description("inproceedings",eng,"Articles in Refereed Conferences").
butype_description("inproceedings",esp,
	"Art\\'{\\i}culos en Congresos con Revisi\\'{o}n Estricta").
%% - Books >with editor<, including formal proceedings
butype_description("book",eng,"Books and Monographs").
butype_description("book",esp,"Libros y Monograf\\'{\\i}as").
%% 2 * Unrefereed, but important publications:
%% -------------------------------------------
%% - Invited *papers* (inc. talks/tutorials when a paper is published).
%%   These have ``inproceedings'' bibtex type but ``invited'' butype. 
butype_description("invited",eng,"Invited Papers and Tutorials").
butype_description("invited",esp,"Art\\'{\\i}ulos Invitados y Tutoriales").
%% - Paper in book >with editor<
butype_description("incollection",eng,
	"Articles in Books and Other Collections").
butype_description("incollection",esp,
	"Art\\'{\\i}culos en libros y otras colecciones").
%% - Workshop papers. These have ``inproceedings'' bibtex
%%   type but ``workshop'' butype.
butype_description("workshop",eng,"Publications in Workshops").
butype_description("workshop",esp,"Publicaciones en Workshops").
butype_description("techreport",eng,"Technical Reports and Manuals").
butype_description("techreport",esp,"Informes T\\'{e}cnicos y Manuales").
%% These are sometimes collapsed to the above:
butype_description("manual",eng,"Technical Reports and Manuals").
butype_description("manual",esp,"Informes T\\'{e}cnicos y Manuales").
butype_description("editedbook",eng,"Edited Books and Proceedings").
butype_description("editedbook",esp,"Libros Editados y Otras Colecciones").
butype_description("mastersthesis",eng,"Master's Theses").
butype_description("mastersthesis",esp,"Tesis de Master").
butype_description("phdthesis",eng,"Ph.D. Theses").
butype_description("phdthesis",esp,"Tesis Doctorales").
%% - No need for publisher, etc. -> e.g., workshop proceedings?:
butype_description("proceedings",eng,"Proceedings").
butype_description("proceedings",esp,"Actas").
%% - Invited talks and tutorials at major conferences, when
%%   there is no paper associated with them (in that case
%%   they are ``invited''). These have ``inproceedings'' bibtex  
%%   type but ``invitedtalk'' butype.
butype_description("invitedtalk",eng,"Invited Talks and Tutorials").
butype_description("invitedtalk",esp,"Charlas Invitadas y Tutoriales").
%% Accepts anything (e.g., internet postings)
butype_description("misc",eng,"Miscellaneous").
butype_description("misc",esp,"Otras Publicaciones").
%% Note required! For manuscripts...
butype_description("unpublished",eng,"Other Documents").
butype_description("unpublished",esp,"Otros Documentos").

:- pop_prolog_flag(multi_arity_warnings).

%-----------------------------------------------------------------------------
:- pred sort_papers(+list,-list,+Type) # "Sort papers. @var{Type}
        determines which fields will be used for the comparison.".
%-----------------------------------------------------------------------------

sort_papers([],[],_Type).
sort_papers([X|L],R,Type) :-
	meta_partition(L,X,L1,L2,Type),
	sort_papers(L1,SL1,Type),
	sort_papers(L2,SL2,Type),
	append(SL1,[X|SL2],R).

meta_partition([],_,[],[],_Type).
meta_partition([X|L],Y,[X|L1],L2,Type) :-
	meta_test(Type,X,Y),
	!,
	meta_partition(L,Y,L1,L2,Type).
meta_partition([X|L],Y,L1,[X|L2],Type) :-
	meta_partition(L,Y,L1,L2,Type).

%-----------------------------------------------------------------------------
:- comment(meta_test/3, "Different comparisons of bib fields in
           papers.  Papers assumed to be in a standard format.").
%-----------------------------------------------------------------------------

:- pred meta_test(^(main_author),Paper1,Paper2) # "Main author of
        @var{Paper1} precedes alphabetically main author of
        @var{Paper2}.".

meta_test(main_author,X,Y) :-
	X=prolog_paper(_KeywordX,ListX),
	Y=prolog_paper(_KeywordY,ListY),
	member_check([author,[MainAX|_]],ListX),
	member_check([author,[MainAY|_]],ListY),
        delete_element(MainAX,T1X,0' ),
        delete_element(MainAY,T1Y,0' ),
        delete_element(T1X,T2X,0'{),
        delete_element(T1Y,T2Y,0'{),
        delete_element(T2X,T3X,0'}),
        delete_element(T2Y,T3Y,0'}),
        % Get rid of initials... (up to 3 >different< ones -- 
        % here is where a regexp would be better)
        replace_list(T3X,T4X,[_,0'.],""),
        replace_list(T4X,T5X,[_,0'.],""),
        replace_list(T5X,AuthXS,[_,0'.],""),
        replace_list(T3Y,T4Y,[_,0'.],""),
        replace_list(T4Y,T5Y,[_,0'.],""),
        replace_list(T5Y,AuthYS,[_,0'.],""),
	% format(user_error,'comparing ~s and ~s ~n',[AuthXS,AuthYS]),
        name(AuthX,AuthXS),
        name(AuthY,AuthYS),
	!,
        AuthX @< AuthY.
	% format(user_error,"TEST SUCCEEDED~n",[]), ttyflush.

:- pred meta_test(^(year_month),Paper1,Paper2) # "Publication date of
        @var{Paper1} precedes that of @var{Paper2}.".

meta_test(year_month,X,Y) :-
	X=prolog_paper(YearXS,MonthXS,_,_),
	Y=prolog_paper(YearYS,MonthYS,_,_),
	[Char]=" ",
	delete_element(YearXS,NYearXS,Char),
	delete_element(YearYS,NYearYS,Char),
	delete_element(MonthXS,NMonthXS,Char),
	delete_element(MonthYS,NMonthYS,Char),
	month_to_number(NMonthXS,NumMonthXS),
	month_to_number(NMonthYS,NumMonthYS),
	append(NYearXS,NumMonthXS,RNYearXS),
	append(NYearYS,NumMonthYS,RNYearYS),
	name(YearX,RNYearXS),
	name(YearY,RNYearYS),
        YearX @< YearY.

month_to_number("January","01"):- !.
month_to_number("February","02"):- !.
month_to_number("March","03"):- !.
month_to_number("April","04"):- !.
month_to_number("May","05"):- !.
month_to_number("June","06"):- !.
month_to_number("July","07"):- !.
month_to_number("August","08"):- !.
month_to_number("September","09"):- !.
month_to_number("October","10"):- !.
month_to_number("November","11"):- !.
month_to_number("December","12"):- !.
month_to_number(_,"00"). % This can apply when month is not known

%-----------------------------------------------------------------------------
% URL database access / handling
%-----------------------------------------------------------------------------

%% add_html2latex_link(_,_,Text,Text).
add_html2latex_link(MAINURL,[],Text,LinkedText) :-
	!,
	do_add_html2latex_link(MAINURL,Text,LinkedText).
add_html2latex_link(MAINURL,AUXURL,Text,TotalLinkedText) :-
	do_add_html2latex_link(MAINURL,Text,LinkedText),
	do_add_html2latex_link(AUXURL,"(also available in ps)",AuxLinkedText),
	append(LinkedText,"\\begin{htmlonly}",T1),
	append(T1,[0'  | AuxLinkedText],T2),
	append(T2,"\\end{htmlonly}",TotalLinkedText).

do_add_html2latex_link(RURL,Text,LinkedText) :-
	replace_element(RURL,URL,0'~,"\\~{}"),
	append("\\htmladdnormallink{",Text,T1),
	append(T1,"}{",T2),
	append(T2,URL,T3),
	append(T3,"}",LinkedText).

add_html2latex_email(EMAIL,Text,TextAndE) :-
	append(Text," \\htmladdnormallink{
\\htmladdimg{http://www.clip.dia.fi.upm.es/images/email\\_blue\\_small.gif}}{",
              T1),
	append(T1,"mailto:",T2),
	append(T2,EMAIL,T3),
	append(T3,"}",TextAndE).

add_url(Author,URLAuthor) :-
	fix_name(Author,PlainName),
	find_url(Author,PlainName,URLAuthor).

:- dynamic url_not_found_reported/1.

find_url(Author,PlainName,URLAuthor) :-
	url_db(PlainName,HTTP,EMAIL),
	!,
        % HTTP is preferred...
  	(  HTTP = "" 
	-> RURL = [0'm,0'a,0'i,0'l,0't,0'o,0':|EMAIL] 
        ;  RURL = HTTP),
	add_html2latex_link(RURL,[],Author,URLAuthor).
%% Uncomment this to add also an email button (ugly):
%%   	(  EMAIL \== "" 
%% 	-> add_html2latex_email(EMAIL,URLAuthor1,URLAuthor)
%%         ;  URLAuthor1 = URLAuthor).
find_url(Author,PlainName,Author) :-
	url_not_found_reported(PlainName), %% Error already reported...
	!.
find_url(Author,PlainName,Author) :-
	assert(url_not_found_reported(PlainName)),
	format(user_error,'[WARNING: URL not found for ~s]~n', [PlainName]).

fix_name(Name,PlainName) :- 
	delete_element(Name,T1,0'\\),
	replace_element(T1,T2,0'~," "),
	delete_element(T2,T3,0'\'),
	delete_element(T3,T4,0'{),
	delete_element(T4,T5,0'}),
	delete_element(T5,T6,0' ),
	delete_element(T6,T7,10), % Newlines...
	delete_element(T7,T8,0'-), 
	delete_element(T8,PlainName,0'}).

:- pred exists_www_copy(Keyword,MAINURL,AUXURL) :: atm * string * string

   # "Checks for the existence of a downloadable (.pdf) file with the
      name @var{Keyword}.pdf (or also @var{Keyword}.ps or
      @var{Keyword}.ps.gz) in a public directory, defined in a fact
      papers_dir_url/2. If a good quality .pdf is available, it
      returns its url in @var{MAINURL} and [] in @var{AUXURL}. If only
      a .ps or .ps.gz is available, it returns its url in
      @var{MAINURL} and [] in @var{AUXURL}. Finally, if only a bad
      quality .pdf is available (flagged by the fact that the file is
      called @var{Keyword}_bitmap.pdf) but @var{Keyword}.ps or
      @var{Keyword}.ps.gz are also available then
      @var{Keyword}_bitmap.pdf is returned in @var{MAINURL} and
      @var{Keyword}.ps or @var{Keyword}.ps.gz is returned also in
      @var{AUXURL}.".

exists_www_copy(Keyword,MAINURL,AUXURL) :-
	papers_dir_url(Dir,DirURL),
	name(Keyword,KeywordS),
	append(Dir,   [0'/ | KeywordS],FileBase),
	append(DirURL,[0'/ | KeywordS],URLBase),
	nofileerrors,
	handle_www_copy_cases(Keyword,FileBase,URLBase,MAINURL,AUXURL),
	fileerrors.
exists_www_copy(Keyword,_MAINURL,_AUXURL) :-
	fileerrors,
	format(user_error,
               "[WARNING: No pdf (nor ps) file found for ~w]~n",[Keyword]),
	fail.

handle_www_copy_cases(_Keyword,FileBase,URLBase,MAINURL,[]) :-
	can_be_opened(FileBase,URLBase,".pdf",MAINURL),
        !.
handle_www_copy_cases(Keyword,FileBase,URLBase,MAINURL,AUXURL) :-
	can_be_opened(FileBase,URLBase,"_bitmap.pdf",MAINURL),
	can_be_opened(FileBase,URLBase,".ps",AUXURL),
	format(user_error,
               "[NOTE: Only bitmap.pdf (and .ps) file found for ~w]~n",
	       [Keyword]),
        !.
handle_www_copy_cases(Keyword,FileBase,URLBase,MAINURL,AUXURL) :-
	can_be_opened(FileBase,URLBase,"_bitmap.pdf",MAINURL),
	can_be_opened(FileBase,URLBase,".ps.gz",AUXURL),
	format(user_error,
               "[NOTE: Only bitmap.pdf (and .ps.gz) file found for ~w]~n",
	       [Keyword]),
        !.
handle_www_copy_cases(Keyword,FileBase,URLBase,MAINURL,[]) :-
	can_be_opened(FileBase,URLBase,"_bitmap.pdf",MAINURL),
	format(user_error,
               "[WARNING: Only _bitmap.pdf file found for ~w]~n",[Keyword]),
        !.
handle_www_copy_cases(Keyword,FileBase,URLBase,MAINURL,[]) :-
	can_be_opened(FileBase,URLBase,".ps",MAINURL),
	format(user_error,
               "[NOTE: Only .ps file found for ~w]~n",[Keyword]),
        !.
handle_www_copy_cases(Keyword,FileBase,URLBase,MAINURL,[]) :-
	can_be_opened(FileBase,URLBase,".ps.gz",MAINURL),
	!,
	format(user_error,
               "[NOTE: Only .ps.gz file found for ~w]~n",[Keyword]).

can_be_opened(FileBase,URLBase,Suffix,URL) :-
	append(FileBase,Suffix,FileNameS),
	name(FileName,FileNameS),
	open(FileName,read,Stream),
	close(Stream),
	append(URLBase,Suffix,URL).
	
:- multifile papers_dir_url/2.
:- dynamic   papers_dir_url/2.

:- pred papers_dir_url(Dir,URL) :: string * string # "@var{Dir} is a
     directory containing .pdf (.ps or .ps.gz) files. @var{URL} is the
     WWW address of that directory.".

%-----------------------------------------------------------------------------
% get beginpage and endpage and/or number of pages
%-----------------------------------------------------------------------------

get_page_number(_,List,Number,In,En) :-
	member_check([beginpage,Init],List), !,
	name(In,Init),
	member_check([endpage,End],List),
	name(En,End),
	Number is (En - In) + 1.
get_page_number(_,List,Number,[],[]) :-
	member_check([npages,Init],List), !,
	name(Number,Init).
get_page_number(K, _L, 'N/A', [], []):-
        format(user_error, 
	       "[WARNING: Page number not available on entry ~w]~n",[K]).

%-----------------------------------------------------------------------------
% Month translation (English/Spanish for now)
%-----------------------------------------------------------------------------

% Backwards compat
spanish_month(Month,TranslatedMonth):- 
	month_to_lang(Month,esp,TranslatedMonth).

month_to_lang("January",esp,"Enero"):- !.
month_to_lang("February",esp,"Febrero"):- !.
month_to_lang("March",esp,"Marzo"):- !.
month_to_lang("April",esp,"Abril"):- !.
month_to_lang("May",esp,"Mayo"):- !.
month_to_lang("June",esp,"Junio"):- !.
month_to_lang("July",esp,"Julio"):- !.
month_to_lang("August",esp,"Agosto"):- !.
month_to_lang("September",esp,"Septiembre"):- !.
month_to_lang("October",esp,"Octubre"):- !.
month_to_lang("November",esp,"Noviembre"):- !.
month_to_lang("December",esp,"Diciembre"):- !.
% Do nothing for English
month_to_lang(X,eng,X):- !.
% Default: do not translate and warn.
month_to_lang(A,L,A) :-
	format(user_error,'[WARNING: Could not translate ~s to language ~w]~n',
                           [A,L]).
	


%-----------------------------------------------------------------------------
:- pred misc_text(Id,Lang,Text) # "@var{Text} is the textual
   description of @var{Id} in language @var{Lang}. ".
%-----------------------------------------------------------------------------

misc_text(volume_short,eng,"Vol.").
misc_text(volume_short,esp,"Vol.").
misc_text(volume_long,eng,"Volume").
misc_text(volume_long,esp,"Vol\\'{u}men").

misc_text(number_short,eng,"Num.").
misc_text(number_short,esp,"N\\'{u}m.").
misc_text(number_long,eng,"Number").
misc_text(number_long,esp,"N\\'{u}mero").

misc_text(inpress,eng,"To Appear").
misc_text(inpress,esp,"En Prensa").

misc_text(pages_short,eng,"Pag.").
misc_text(pages_short,esp,"P\\'{a}g.").
misc_text(pages_long,eng,"pages").
misc_text(pages_long,esp,"p\\'{a}ginas").

misc_text(msthesis,eng,"Ms.\\ Thesis").
misc_text(msthesis,esp,"Tesis de Licenciatura").

misc_text(phdthesis,eng,"Ph.D.\\ Thesis").
misc_text(phdthesis,esp,"Tesis Doctoral").

%-----------------------------------------------------------------------------
% String x List of Strings
%-----------------------------------------------------------------------------

is_an_author(Name,[String|_Rest]) :-
%%	sublist(Name,String), %% Wrong!
	list_contained(Name,String),
	!.
is_an_author(Name,[_String|Rest]) :-
	is_an_author(Name,Rest).

%-----------------------------------------------------------------------------
% Miscellaneous other support predicates
%-----------------------------------------------------------------------------

% InputList x OutputList x InputSubListA x InputSublistB
% OutputList is InputList with InputSublistA replaced by InputSublistB
% (ideally, we should add regexps)
replace_list([],[],_,_).
replace_list(IL,OL,X,Y) :-
	X = [_|_],
        prefix_rest(X,IL,RIL),
	!,
	replace_list(RIL,NOL,X,Y),
        append(Y,NOL,OL).
replace_list([Z|L],[Z|NL],X,Y) :-
        replace_list(L,NL,X,Y).

% InputPrefix x InputList x List
% InputPrefix is a prefix of InputList and List is the rest.
prefix_rest([], R, R).
prefix_rest([X|PT], [X|T], R) :-
	prefix_rest(PT, T, R).

% ... Now, several specializations of the above...

% InputList x OutputList x Char
delete_element([],[],_).
delete_element([X|L],NL,X) :- 
	!,
	delete_element(L,NL,X).
delete_element([Y|L],[Y|NL],X) :- 
	delete_element(L,NL,X).

% InputList x OutputList x Char x List
replace_element([],[],_,_).
replace_element([X|L],R,NX,Y) :- 
	X==NX,
	!,
	replace_element(L,NL,X,Y),
	append(Y,NL,R).
replace_element([Z|L],[Z|NL],X,Y) :- 
	replace_element(L,NL,X,Y).

% ISO escaping...
% *** change to use replace element
prolog_escape([],[]).
prolog_escape([0'\\|T],[0'\\,0'\\|NT]) :- 
	!,
	prolog_escape(T,NT).
prolog_escape([0'"|T],[0'\\,0'"|NT]) :- 
	!,
	prolog_escape(T,NT).
prolog_escape([Y|L],[Y|NL]) :- 
	prolog_escape(L,NL).

delete_spaces([],[]).
delete_spaces([9|R],RD) :-
	!,
	delete_spaces(R,RD).
delete_spaces([10|R],RD) :-
	!,
	delete_spaces(R,RD).
delete_spaces([32|R],RD) :-
	!,
	delete_spaces(R,RD).
delete_spaces([X|R],[X|RD]) :-
	!,
	delete_spaces(R,RD).

% 32 is space
delete_extra_spaces([],[]).
delete_extra_spaces([32,32|T],NT) :-
	!,
	delete_extra_spaces([32|T],NT).
delete_extra_spaces([X|T],[X|NT]) :-
	delete_extra_spaces(T,NT).

list_contained(List1, List2) :-
	append(X,_,List2),
	append(_,List1,X).

sublist(List, List).
sublist(Sub, [Head|Tail]) :- sublist_(Tail, Head, Sub).

sublist_(Sub, _, Sub).
sublist_([Head|Tail], _, Sub) :- sublist_(Tail, Head, Sub).
sublist_([Head|Tail], X, [X|Sub]) :- sublist_(Tail, Head, Sub).

nreverse([],[]).
nreverse([X|L0],L) :- 
	nreverse(L0,L1), 
	append(L1,[X],L).

append([],L,L).
append([X|L1],L2,[X|L3]) :- 
	append(L1,L2,L3).

% Watch out: member_check, really!
member_check(X,[X|_]) :- !.
member_check(X,[_|T]) :- member_check(X,T).

write_strlist([],_Separator,End,Stream) :-
	format(Stream,"~s",End).
write_strlist([Y|RY],Separator,End,Stream) :-
	format(Stream,"~s~s",[Y,Separator]),
	write_strlist(RY,Separator,End,Stream).

to_lower_case([],[]).
to_lower_case([H|T],[NH|NT]) :-
	is_upper_case(H),
	!,
	NH is H+32,
	to_lower_case(T,NT).
to_lower_case([H|T],[H|NT]) :-
	to_lower_case(T,NT).

add_capital_bracing([],[]).
add_capital_bracing([C|T],[0'{,C|NT]) :-
	is_upper_case(C),
	!,
	continue_uppercase(T,NT).
add_capital_bracing([C|T],[C|NT]) :-
	add_capital_bracing(T,NT).


continue_uppercase([],[0'}]).
continue_uppercase([C|T],[125,C|NT]) :-
	is_lower_case(C),
	!,
	add_capital_bracing(T,NT).
continue_uppercase([C|T],[C|NT]) :-
	continue_uppercase(T,NT).

is_upper_case(Ch) :-
	Ch >= 65, Ch =< 90.

is_lower_case(Ch) :-
	Ch >= 97, Ch =< 122.

%----------------------------------------------------------------------
:- comment(version_maintenance,dir(version)).

:- comment(version(0*4+63,2004/04/05,16:07*26+'CEST'), "Fixed minor
   typo in English translation of Invited Papers and Tutorials.
   (Manuel Hermenegildo)").

:- comment(version(0*4+54,2004/02/28,23:46*40+'CET'), "Now 'also
   available in ps' does not appear in the tex/ps/pdf versions.
   (Manuel Hermenegildo)").

:- comment(version(0*4+40,2002/04/06,19:54*21+'CEST'), "Changes for
   some multi-lingual support (eng/esp).  (Manuel Hermenegildo)").

:- comment(version(0*4+38,2001/11/18,22:47*54+'CET'), "Minor change to
   how incollection is listed.  (Manuel Hermenegildo)").

:- comment(version(0*4+35,1999/09/02,21:45*57+'MEST'), "@tt{\"}
   characters within bib entry fields now supported, as needed for
   umlaut, etc.  (Manuel Hermenegildo)").

:- comment(version(0*4+21,1999/03/16,14:22*35+'MET'), "Changed
   syntax/1 to use_package/1 for 0.9.  (Manuel Hermenegildo)").

:- comment(version(0*4+16,1998/10/27,12:25*08+'EST'), "Errors on urls
   not found now given only once per run.  (Manuel Hermenegildo)").

:- comment(version(0*4+4,1998/10/18,13:22*19+'MET DST'), "Now running
   natively in CIAO Prolog. Added some more comments. (Manuel
   Hermenegildo)").
%----------------------------------------------------------------------
